@@ -22,6 +22,8 @@
 
 #include "gitg-diff-view.h"
 #include "gitg-types.h"
+#include "gitg-diff-line-renderer.h"
+
 #include <string.h>
 #include <stdlib.h>
 
@@ -40,6 +42,17 @@
 static void on_buffer_insert_text(GtkTextBuffer *buffer, GtkTextIter *iter, gchar const *text, gint len, GitgDiffView *view);
 static void on_buffer_delete_range(GtkTextBuffer *buffer, GtkTextIter *start, GtkTextIter *end, GitgDiffView *view);
 
+static void
+line_renderer_size_func (GtkSourceGutter *gutter,
+                         GtkCellRenderer *cell,
+                         GitgDiffView    *view);
+static void
+line_renderer_data_func (GtkSourceGutter *gutter,
+                         GtkCellRenderer *cell,
+                         gint             line_number,
+                         gboolean         current_line,
+                         GitgDiffView    *view);
+
 static gboolean on_idle_scan(GitgDiffView *view);
 
 /* Signals */
@@ -54,7 +67,7 @@ enum
 enum
 {
 	PROP_0,
-	
+
 	PROP_DIFF_ENABLED
 };
 
@@ -93,11 +106,18 @@ struct _GitgDiffViewPrivate
 	Region *regions;
 	Region *last_region;
 	GSequence *regions_index;
-	
+
 	guint scan_id;
 	gboolean diff_enabled;
 	GtkTextBuffer *current_buffer;
 	GtkTextTag *invisible_tag;
+	GtkTextTag *subheader_tag;
+
+	GitgDiffLineRenderer *line_renderer;
+
+	Region *lines_current_region;
+	gint lines_previous_line;
+	guint lines_counters[2];
 };
 
 G_DEFINE_TYPE(GitgDiffView, gitg_diff_view, GTK_TYPE_SOURCE_VIEW)
@@ -110,9 +130,9 @@ region_free(Region *region)
 {
 	if (!region)
 		return;
-	
+
 	region_free(region->next);
-	
+
 	if (region->type == GITG_DIFF_ITER_TYPE_HEADER)
 		g_slice_free(Header, (Header *)region);
 	else
@@ -124,7 +144,7 @@ regions_free(GitgDiffView *view, gboolean remove_signals)
 {
 	region_free(view->priv->regions);
 	g_sequence_remove_range(g_sequence_get_begin_iter(view->priv->regions_index), g_sequence_get_end_iter(view->priv->regions_index));
-	
+
 	view->priv->regions = NULL;
 	view->priv->last_region = NULL;
 	view->priv->last_scan_line = 0;
@@ -135,30 +155,32 @@ regions_free(GitgDiffView *view, gboolean remove_signals)
 		g_source_remove(view->priv->scan_id);
 		view->priv->scan_id = 0;
 	}
-	
+
 	if (view->priv->current_buffer && remove_signals)
 	{
 		g_signal_handlers_disconnect_by_func(view->priv->current_buffer, G_CALLBACK(on_buffer_insert_text), view);
 		g_signal_handlers_disconnect_by_func(view->priv->current_buffer, G_CALLBACK(on_buffer_delete_range), view);
 
-		gtk_text_tag_table_remove(gtk_text_buffer_get_tag_table(view->priv->current_buffer), view->priv->invisible_tag);
-		
-		g_object_unref(view->priv->current_buffer);
-		
+		gtk_text_tag_table_remove (gtk_text_buffer_get_tag_table(view->priv->current_buffer), view->priv->invisible_tag);
+		gtk_text_tag_table_remove (gtk_text_buffer_get_tag_table(view->priv->current_buffer), view->priv->subheader_tag);
+
+		g_object_unref (view->priv->current_buffer);
+
 		view->priv->current_buffer = NULL;
 		view->priv->invisible_tag = NULL;
+		view->priv->subheader_tag = NULL;
 	}
 }
-							 
+
 static void
-gitg_diff_view_finalize(GObject *object)
+gitg_diff_view_finalize (GObject *object)
 {
-	GitgDiffView *view = GITG_DIFF_VIEW(object);
-	
-	regions_free(view, TRUE);
-	g_sequence_free(view->priv->regions_index);
-	
-	G_OBJECT_CLASS(gitg_diff_view_parent_class)->finalize(object);
+	GitgDiffView *view = GITG_DIFF_VIEW (object);
+
+	regions_free (view, TRUE);
+	g_sequence_free (view->priv->regions_index);
+
+	G_OBJECT_CLASS (gitg_diff_view_parent_class)->finalize (object);
 }
 
 static void
@@ -172,7 +194,7 @@ static void
 gitg_diff_view_set_property(GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
 {
 	GitgDiffView *self = GITG_DIFF_VIEW(object);
-	
+
 	switch (prop_id)
 	{
 		case PROP_DIFF_ENABLED:
@@ -200,16 +222,24 @@ gitg_diff_view_get_property(GObject *object, guint prop_id, GValue *value, GPara
 	}
 }
 
+static void
+gitg_diff_view_constructed (GObject *object)
+{
+	g_object_set (object, "show-line-numbers", FALSE, NULL);
+}
+
 static void
 gitg_diff_view_class_init(GitgDiffViewClass *klass)
 {
 	GObjectClass *object_class = G_OBJECT_CLASS(klass);
 	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS(klass);
-	
+
 	object_class->finalize = gitg_diff_view_finalize;
 	object_class->set_property = gitg_diff_view_set_property;
 	object_class->get_property = gitg_diff_view_get_property;
 
+	object_class->constructed = gitg_diff_view_constructed;
+
 	widget_class->expose_event = gitg_diff_view_expose;
 
 	diff_view_signals[HEADER_ADDED] =
@@ -249,98 +279,56 @@ on_buffer_set(GitgDiffView *self, GParamSpec *spec, gpointer userdata)
 {
 	/* remove all regions for a new buffer */
 	regions_free(self, TRUE);
-	
+
 	self->priv->current_buffer = g_object_ref(gtk_text_view_get_buffer(GTK_TEXT_VIEW(self)));
 	g_signal_connect_after(self->priv->current_buffer, "insert-text", G_CALLBACK(on_buffer_insert_text), self);
 	g_signal_connect_after(self->priv->current_buffer, "delete-range", G_CALLBACK(on_buffer_delete_range), self);
 
 	self->priv->scan_id = g_idle_add((GSourceFunc)on_idle_scan, self);
 	self->priv->invisible_tag = gtk_text_buffer_create_tag(self->priv->current_buffer, "GitgHunkInvisible", "invisible", TRUE, NULL);
+	self->priv->subheader_tag = gtk_text_buffer_create_tag(self->priv->current_buffer, "GitgHunkSubHeader", "invisible", TRUE, NULL);
+
+	gtk_text_tag_set_priority (self->priv->subheader_tag,
+	                           gtk_text_tag_table_get_size (gtk_text_buffer_get_tag_table (self->priv->current_buffer)) - 1);
 }
 
 static void
-gitg_diff_view_init(GitgDiffView *self)
+gitg_diff_view_init (GitgDiffView *self)
 {
-	self->priv = GITG_DIFF_VIEW_GET_PRIVATE(self);
-	
-	self->priv->regions_index = g_sequence_new(NULL);
-	
-	g_signal_connect(self, "notify::buffer", G_CALLBACK(on_buffer_set), NULL);
+	GtkSourceGutter *gutter;
+
+	self->priv = GITG_DIFF_VIEW_GET_PRIVATE (self);
+
+	self->priv->regions_index = g_sequence_new (NULL);
+	self->priv->line_renderer = gitg_diff_line_renderer_new ();
+
+	g_object_set (self->priv->line_renderer, "xpad", 2, NULL);
+
+	gutter = gtk_source_view_get_gutter (GTK_SOURCE_VIEW (self), GTK_TEXT_WINDOW_LEFT);
+
+	gtk_source_gutter_insert (gutter,
+	                          GTK_CELL_RENDERER (self->priv->line_renderer),
+	                          0);
+
+	gtk_source_gutter_set_cell_data_func (gutter,
+	                                      GTK_CELL_RENDERER (self->priv->line_renderer),
+	                                      (GtkSourceGutterDataFunc)line_renderer_data_func,
+	                                      self,
+	                                      NULL);
+
+	gtk_source_gutter_set_cell_size_func (gutter,
+	                                      GTK_CELL_RENDERER (self->priv->line_renderer),
+	                                      (GtkSourceGutterSizeFunc)line_renderer_size_func,
+	                                      self,
+	                                      NULL);
+
+	g_signal_connect (self, "notify::buffer", G_CALLBACK (on_buffer_set), NULL);
 }
 
-GitgDiffView*
-gitg_diff_view_new()
+GitgDiffView *
+gitg_diff_view_new ()
 {
-	return g_object_new(GITG_TYPE_DIFF_VIEW, NULL);
-}
-
-/* This function is taken from gtk+/tests/testtext.c */
-static void
-get_lines (GtkTextView *text_view, gint first_y, gint last_y, GArray *buffer_coords, GArray *line_heights, GArray *numbers, gint *countp)
-{
-	GtkTextIter iter;
-	gint count;
-	gint size;
-	gint last_line_num = -1;
-
-	g_array_set_size(buffer_coords, 0);
-	g_array_set_size(numbers, 0);
-	
-	if (line_heights != NULL)
-		g_array_set_size(line_heights, 0);
-
-	/* Get iter at first y */
-	gtk_text_view_get_line_at_y(text_view, &iter, first_y, NULL);
-
-	/* For each iter, get its location and add it to the arrays.
-	 * Stop when we pass last_y */
-	count = 0;
-  	size = 0;
-
-  	while (!gtk_text_iter_is_end(&iter))
-    {
-		gint y, height;
-
-		gtk_text_view_get_line_yrange(text_view, &iter, &y, &height);
-
-		g_array_append_val(buffer_coords, y);
-		
-		if (line_heights)
-			g_array_append_val(line_heights, height);
-			
-		last_line_num = gtk_text_iter_get_line(&iter);
-		g_array_append_val(numbers, last_line_num);
-
-		++count;
-
-		if ((y + height) >= last_y)
-			break;
-
-		gtk_text_iter_forward_line(&iter);
-	}
-
-	if (gtk_text_iter_is_end(&iter))
-    {
-		gint y, height;
-		gint line_num;
-
-		gtk_text_view_get_line_yrange(text_view, &iter, &y, &height);
-
-		line_num = gtk_text_iter_get_line(&iter);
-
-		if (line_num != last_line_num)
-		{
-			g_array_append_val(buffer_coords, y);
-			
-			if (line_heights)
-				g_array_append_val(line_heights, height);
-
-			g_array_append_val(numbers, line_num);
-			++count;
-		}
-	}
-
-	*countp = count;
+	return g_object_new (GITG_TYPE_DIFF_VIEW, NULL);
 }
 
 static gint
@@ -348,18 +336,41 @@ index_compare(gconstpointer a, gconstpointer b, gpointer userdata)
 {
 	guint la = ((Region *)a)->line;
 	guint lb = ((Region *)b)->line;
-	
+
 	return la < lb ? -1 : (la > lb ? 1 : 0);
 }
 
 static void
-ensure_max_line(GitgDiffView *view, Hunk *hunk)
+ensure_max_line (GitgDiffView *view, Hunk *hunk)
 {
 	guint num = hunk->region.next ? hunk->region.next->line - hunk->region.line : 0;
-	guint m = MAX(hunk->new + num, hunk->old + num);
+	guint m = MAX (hunk->new + num, hunk->old + num);
 
 	if (m > view->priv->max_line_count)
+	{
 		view->priv->max_line_count = m;
+
+		gtk_source_gutter_queue_draw (gtk_source_view_get_gutter (GTK_SOURCE_VIEW (view), GTK_TEXT_WINDOW_LEFT));
+	}
+}
+
+static void
+hide_header_details (GitgDiffView *view,
+                     Region       *region)
+{
+	/* Just hide the lines 2->n lines from region to region->next */
+	GtkTextBuffer *buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (view));
+	GtkTextIter startiter;
+	GtkTextIter enditer;
+
+	gint line = region->line + 2;
+
+	gtk_text_buffer_get_iter_at_line (buffer, &startiter, line);
+
+	enditer = startiter;
+	gtk_text_iter_forward_lines (&enditer, region->next->line - line);
+
+	gtk_text_buffer_apply_tag (buffer, view->priv->subheader_tag, &startiter, &enditer);
 }
 
 static void
@@ -369,9 +380,11 @@ add_region(GitgDiffView *view, Region *region)
 	{
 		view->priv->last_region->next = region;
 		region->prev = view->priv->last_region;
-		
+
 		if (view->priv->last_region->type == GITG_DIFF_ITER_TYPE_HUNK)
+		{
 			ensure_max_line(view, (Hunk *)view->priv->last_region);
+		}
 	}
 	else
 	{
@@ -380,35 +393,48 @@ add_region(GitgDiffView *view, Region *region)
 	}
 
 	view->priv->last_region = region;
+
+	if (region->prev && region->prev->type == GITG_DIFF_ITER_TYPE_HEADER)
+	{
+		/* Hide header details if first hunk is scanned */
+		hide_header_details (view, region->prev);
+	}
+
 	g_sequence_insert_sorted(view->priv->regions_index, region, index_compare, NULL);
-	
+
 	GitgDiffIter iter;
 	iter.userdata = view;
 	iter.userdata2 = region;
-	
+
 	if (region->type == GITG_DIFF_ITER_TYPE_HEADER)
+	{
 		g_signal_emit(view, diff_view_signals[HEADER_ADDED], 0, &iter);
+	}
 	else if (region->type == GITG_DIFF_ITER_TYPE_HUNK)
+	{
 		g_signal_emit(view, diff_view_signals[HUNK_ADDED], 0, &iter);
+	}
 }
 
 static void
 parse_hunk_info(Hunk *hunk, GtkTextIter *iter)
 {
 	GtkTextIter end = *iter;
-	
+
 	gtk_text_iter_forward_to_line_end(&end);
 	gchar *text = gtk_text_iter_get_text(iter, &end);
-	
+
 	hunk->old = 0;
 	hunk->new = 0;
-	
+
 	gchar *old = g_utf8_strchr(text, -1, '-');
 	gchar *new = g_utf8_strchr(text, -1, '+');
 
 	if (!old || !new)
+	{
 		return;
-	
+	}
+
 	hunk->old = atoi(old + 1);
 	hunk->new = atoi(new + 1);
 
@@ -422,22 +448,26 @@ ensure_scan(GitgDiffView *view, guint last_line)
 	GtkTextIter iter;
 	GtkTextBuffer *buffer = view->priv->current_buffer;
 	gtk_text_buffer_get_iter_at_line(buffer, &iter, view->priv->last_scan_line);
-	
+
 	while (view->priv->last_scan_line <= last_line)
 	{
 		GtkTextIter start = iter;
 		GtkTextIter end = iter;
 
 		if (!gtk_text_iter_forward_line(&iter))
+		{
 			break;
+		}
 
 		++view->priv->last_scan_line;
-		
+
 		if (!gtk_text_iter_forward_chars(&end, 3))
+		{
 			continue;
+		}
 
 		gchar *text = gtk_text_iter_get_text(&start, &end);
-		
+
 		if (g_str_has_prefix(text, "@@ "))
 		{
 			/* start new hunk region */
@@ -445,22 +475,24 @@ ensure_scan(GitgDiffView *view, guint last_line)
 			hunk->region.type = GITG_DIFF_ITER_TYPE_HUNK;
 			hunk->region.line = view->priv->last_scan_line - 1;
 			hunk->region.visible = TRUE;
-			
+
 			parse_hunk_info(hunk, &start);
-			
+
 			add_region(view, (Region *)hunk);
-			
+
 			g_free(text);
 			continue;
 		}
-		
+
 		g_free(text);
 
 		if (!gtk_text_iter_forward_chars(&end, 7))
+		{
 			continue;
-		
+		}
+
 		text = gtk_text_iter_get_text(&start, &end);
-		
+
 		if (g_str_has_prefix(text, "diff --git") || g_str_has_prefix(text, "diff --cc"))
 		{
 			/* start new header region */
@@ -468,18 +500,20 @@ ensure_scan(GitgDiffView *view, guint last_line)
 			header->region.type = GITG_DIFF_ITER_TYPE_HEADER;
 			header->region.line = view->priv->last_scan_line - 1;
 			header->region.visible = TRUE;
-			
+
 			header->index_to[0] = '\0';
 			header->index_from[0] = '\0';
-			
+
 			add_region(view, (Region *)header);
 		}
-		
+
 		g_free(text);
 	}
-	
+
 	if (view->priv->last_region && view->priv->last_region->type == GITG_DIFF_ITER_TYPE_HUNK)
+	{
 		ensure_max_line(view, (Hunk *)view->priv->last_region);
+	}
 }
 
 static Region *
@@ -487,20 +521,24 @@ find_current_region(GitgDiffView *view, guint line)
 {
 	GSequenceIter *iter;
 	Region tmp = {0, NULL, NULL, line};
-	
+
 	iter = g_sequence_search(view->priv->regions_index, &tmp, index_compare, NULL);
-	
+
 	if (!iter || g_sequence_iter_is_begin(iter))
+	{
 		return NULL;
+	}
 
 	if (!g_sequence_iter_is_end(iter))
 	{
 		Region *ret = (Region *)g_sequence_get(iter); 
-	
+
 		if (ret->line == line)
+		{
 			return ret->visible ? ret : NULL;
+		}
 	}
-	 
+
 	Region *ret = (Region *)g_sequence_get(g_sequence_iter_prev(iter));
 	return ret->visible ? ret : NULL;
 }
@@ -509,17 +547,20 @@ static gboolean
 line_has_prefix(GitgDiffView *view, guint line, gchar const *prefix)
 {
 	GtkTextIter iter;
-	
+
 	gtk_text_buffer_get_iter_at_line(view->priv->current_buffer, &iter, line);
 
 	GtkTextIter end = iter;
+
 	if (!gtk_text_iter_forward_chars(&end, g_utf8_strlen(prefix, -1)))
+	{
 		return FALSE;
-	
+	}
+
 	gchar *text = gtk_text_iter_get_text(&iter, &end);
 	gboolean ret = g_str_has_prefix(text, prefix);
 	g_free(text);
-	
+
 	return ret;
 }
 
@@ -539,160 +580,104 @@ static void
 get_initial_counters(GitgDiffView *view, Region *region, guint line, guint counters[2])
 {
 	guint i;
-	
+
 	counters[0] = counters[1] = 0;
-	
+
 	for (i = region->line + 1; i < line; ++i)
 	{
 		if (draw_old(view, i))
 			++counters[0];
-		
+
 		if (draw_new(view, i))
 			++counters[1];
 	}
 }
 
 static void
-paint_line_numbers(GitgDiffView *view, GdkEventExpose *event)
+line_renderer_size_func (GtkSourceGutter *gutter,
+                         GtkCellRenderer *cell,
+                         GitgDiffView    *view)
 {
-	GtkTextView *text_view;
-	GdkWindow *win;
-	PangoLayout *layout;
-	GArray *numbers;
-	GArray *pixels;
-	gchar str_old[16];  /* we don't expect more than ten million lines ;-) */
-	gchar str_new[16];
-	gint y1, y2;
-	gint count;
-	gint margin_width;
-	gint text_width;
-	gint i;
-	GtkTextIter cur;
-
-	text_view = GTK_TEXT_VIEW(view);
-	win = gtk_text_view_get_window(text_view, GTK_TEXT_WINDOW_LEFT);
-
-	y1 = event->area.y;
-	y2 = y1 + event->area.height;
-
-	/* get the extents of the line printing */
-	gtk_text_view_window_to_buffer_coords(text_view, GTK_TEXT_WINDOW_LEFT, 0, y1, NULL, &y1);
-	gtk_text_view_window_to_buffer_coords(text_view, GTK_TEXT_WINDOW_LEFT, 0, y2, NULL, &y2);
-
-	numbers = g_array_new(FALSE, FALSE, sizeof(gint));
-	pixels = g_array_new(FALSE, FALSE, sizeof(gint));
-
-	/* get the line numbers and y coordinates. */
-	get_lines(text_view, y1, y2, pixels, NULL, numbers, &count);
-	
-	/* A zero-lined document should display a "1"; we don't need to worry about
-	scrolling effects of the text widget in this special case */
-
-	if (count == 0)
-	{
-		gint y = 0;
-		gint n = 0;
-		count = 1;
-		g_array_append_val(pixels, y);
-		g_array_append_val(numbers, n);
-	}
-	
-	/* Ensure scanned until last needed line */	
-	guint last = g_array_index(numbers, gint, count - 1);
-	ensure_scan(view, last);
-
-	/* set size */
-	g_snprintf(str_old, sizeof(str_old), "%d", MAX(99, view->priv->max_line_count));
-	layout = gtk_widget_create_pango_layout(GTK_WIDGET(view), str_old);
-
-	pango_layout_get_pixel_size(layout, &text_width, NULL);
-
-	/* determine the width of the left margin. */
-	margin_width = text_width * 2 + 9;
-	
-	guint extra_width = 0;
-	
-	if (gtk_source_view_get_show_line_marks(GTK_SOURCE_VIEW(view)))
-		extra_width = 20;
-
-	pango_layout_set_width(layout, text_width);
-	pango_layout_set_alignment(layout, PANGO_ALIGN_RIGHT);
-
-	gtk_text_view_set_border_window_size(GTK_TEXT_VIEW(text_view), GTK_TEXT_WINDOW_LEFT, margin_width + extra_width);
-	gtk_text_buffer_get_iter_at_mark(text_view->buffer, &cur, gtk_text_buffer_get_insert(text_view->buffer));
-
-	Region *current = NULL;
-	guint counters[2];
-
-	for (i = 0; i < count; ++i)
-	{
-		gint pos;
-		gint line_to_paint;
-
-		gtk_text_view_buffer_to_window_coords(text_view, GTK_TEXT_WINDOW_LEFT, 0, g_array_index(pixels, gint, i), NULL, &pos);
-		line_to_paint = g_array_index(numbers, gint, i);
-		
-		if (!current)
-		{
-			current = find_current_region(view, line_to_paint);
-			
-			if (current)
-				get_initial_counters(view, current, line_to_paint, counters);
-		}
-		
-		*str_old = '\0';
-		*str_new = '\0';
-
-		if (current && current->type == GITG_DIFF_ITER_TYPE_HUNK && line_to_paint != current->line)
-		{
-			Hunk *hunk = (Hunk *)current;
-			
-			if (draw_old(view, line_to_paint))
-				g_snprintf(str_old, sizeof(str_old), "%d", hunk->old + counters[0]++);
-
-			if (draw_new(view, line_to_paint))
-				g_snprintf(str_new, sizeof(str_new), "%d", hunk->new + counters[1]++);
-		}
-		
-		pango_layout_set_markup(layout, str_old, -1);
-		gtk_paint_layout(GTK_WIDGET(view)->style, win, GTK_WIDGET_STATE(view), FALSE, NULL, GTK_WIDGET(view), NULL, margin_width - 7 - text_width, pos, layout);
-
-		pango_layout_set_markup(layout, str_new, -1);
-		gtk_paint_layout(GTK_WIDGET(view)->style, win, GTK_WIDGET_STATE(view), FALSE, NULL, GTK_WIDGET(view), NULL, margin_width - 2, pos, layout);
-
-		if (current && current->next && line_to_paint == current->next->line - 1)
-		{
-			counters[0] = counters[1] = 0;
-			current = current->next->visible ? current->next : NULL;
-		}
-	}
-	
-	gtk_paint_vline(GTK_WIDGET(view)->style, win, GTK_WIDGET_STATE(view), NULL, GTK_WIDGET(view), NULL, event->area.y, event->area.y + event->area.height, 4 + text_width);
-
-	g_array_free(pixels, TRUE);
-	g_array_free(numbers, TRUE);
-
-	g_object_unref(G_OBJECT(layout));
+	g_object_set (cell,
+	              "line_old", view->priv->max_line_count,
+	              "line_new", view->priv->max_line_count,
+	              NULL);
 }
 
-static gint 
-gitg_diff_view_expose(GtkWidget *widget, GdkEventExpose *event)
+static void
+line_renderer_data_func (GtkSourceGutter *gutter,
+                         GtkCellRenderer *cell,
+                         gint             line_number,
+                         gboolean         current_line,
+                         GitgDiffView    *view)
 {
-	gboolean ret = FALSE;
-	GtkTextView *text_view = GTK_TEXT_VIEW(widget);
-	GitgDiffView *view = GITG_DIFF_VIEW(widget);
+	gint line_old = -1;
+	gint line_new = -1;
+	Region **current = &view->priv->lines_current_region;
 
-	if (event->window == gtk_text_view_get_window(text_view, GTK_TEXT_WINDOW_LEFT) && 
-	    view->priv->diff_enabled && gtk_source_view_get_show_line_numbers(GTK_SOURCE_VIEW(view)))
+	ensure_scan (view, line_number);
+
+	if (!*current || view->priv->lines_previous_line + 1 != line_number)
 	{
-		paint_line_numbers(GITG_DIFF_VIEW(widget), event);
-		ret = TRUE;
+		*current = find_current_region (view, line_number);
+
+		if (*current)
+		{
+			get_initial_counters (view,
+			                      *current,
+			                      line_number,
+			                      view->priv->lines_counters);
+		}
 	}
 
-	if (GTK_WIDGET_CLASS(gitg_diff_view_parent_class)->expose_event)
-		ret = ret || GTK_WIDGET_CLASS(gitg_diff_view_parent_class)->expose_event(widget, event);
+	view->priv->lines_previous_line = line_number;
 
-	return ret;
+	if (*current &&
+	    (*current)->type == GITG_DIFF_ITER_TYPE_HUNK &&
+	    line_number != (*current)->line)
+	{
+		Hunk *hunk = (Hunk *)*current;
+
+		if (draw_old (view, line_number))
+		{
+			line_old = hunk->old + view->priv->lines_counters[0]++;
+		}
+
+		if (draw_new (view, line_number))
+		{
+			line_new = hunk->new + view->priv->lines_counters[1]++;
+		}
+	}
+
+	g_object_set (cell, "line_old", line_old, "line_new", line_new, NULL);
+
+	if (*current && (*current)->next && line_number == (*current)->next->line - 1)
+	{
+		view->priv->lines_counters[0] = view->priv->lines_counters[1] = 0;
+		*current = (*current)->next->visible ? (*current)->next : NULL;
+	}
+}
+
+static gint
+gitg_diff_view_expose (GtkWidget      *widget,
+                       GdkEventExpose *event)
+{
+	GitgDiffView *view = GITG_DIFF_VIEW (widget);
+
+	/* Prepare for new round of expose on the line renderer */
+	view->priv->lines_current_region = NULL;
+	view->priv->lines_previous_line = -1;
+	view->priv->lines_counters[0] = 0;
+	view->priv->lines_counters[1] = 0;
+
+	if (GTK_WIDGET_CLASS (gitg_diff_view_parent_class)->expose_event)
+	{
+		return GTK_WIDGET_CLASS (gitg_diff_view_parent_class)->expose_event (widget, event);
+	}
+	else
+	{
+		return FALSE;
+	}
 }
 
 void
@@ -709,19 +694,19 @@ void
 gitg_diff_view_remove_hunk(GitgDiffView *view, GtkTextIter *iter)
 {
 	g_return_if_fail(GITG_IS_DIFF_VIEW(view));
-	
+
 	/* removes hunk at iter and if it was the last hunk of a file, also removes
 	   the file header */
 	Region *region = find_current_region(view, gtk_text_iter_get_line(iter));
-	
+
 	if (!region)
 		return;
-	
+
 	GtkTextIter start;
 	GtkTextIter end;
-	
+
 	gtk_text_buffer_get_iter_at_line(view->priv->current_buffer, &start, region->line);
-	
+
 	if (region->next)
 	{
 		gtk_text_buffer_get_iter_at_line(view->priv->current_buffer, &end, region->next->line - 1);
@@ -731,9 +716,9 @@ gitg_diff_view_remove_hunk(GitgDiffView *view, GtkTextIter *iter)
 	{
 		gtk_text_buffer_get_end_iter(view->priv->current_buffer, &end);
 	}
-	
+
 	Region *prev = find_current_region(view, region->line - 1);
-	
+
 	if ((!region->next || region->next->type == GITG_DIFF_ITER_TYPE_HEADER) && (!prev || prev->type == GITG_DIFF_ITER_TYPE_HEADER))
 	{
 		if (!prev)
@@ -741,7 +726,7 @@ gitg_diff_view_remove_hunk(GitgDiffView *view, GtkTextIter *iter)
 		else
 			gtk_text_buffer_get_iter_at_line(view->priv->current_buffer, &start, region->line);
 	}
-	
+
 	gtk_text_buffer_delete(view->priv->current_buffer, &start, &end);
 }
 
@@ -753,7 +738,7 @@ gitg_diff_view_get_start_iter(GitgDiffView *view, GitgDiffIter *iter)
 
 	iter->userdata = view;
 	iter->userdata2 = view->priv->regions;
-	
+
 	return iter->userdata2 != NULL;
 }
 
@@ -761,10 +746,10 @@ gboolean
 gitg_diff_iter_forward(GitgDiffIter *iter)
 {
 	g_return_val_if_fail(iter != NULL, FALSE);
-	
+
 	if (!iter->userdata2)
 		return FALSE;
-	
+
 	iter->userdata2 = ((Region *)iter->userdata2)->next;
 
 	return iter->userdata2 != NULL;
@@ -778,7 +763,7 @@ gitg_diff_view_get_end_iter(GitgDiffView *view, GitgDiffIter *iter)
 
 	iter->userdata = view;
 	iter->userdata2 = view->priv->last_region;
-	
+
 	return iter->userdata2 != NULL;
 }
 
@@ -786,14 +771,14 @@ gboolean
 gitg_diff_iter_backward(GitgDiffIter *iter)
 {
 	g_return_val_if_fail(iter != NULL, FALSE);
-	
+
 	if (!iter->userdata2)
 		return FALSE;
-	
+
 	iter->userdata2 = ((Region *)iter->userdata2)->prev;
 
 	return iter->userdata2 != NULL;
-	
+
 }
 
 GitgDiffIterType 
@@ -802,7 +787,7 @@ gitg_diff_iter_get_type(GitgDiffIter *iter)
 	g_return_val_if_fail(iter != NULL, 0);
 	g_return_val_if_fail(GITG_IS_DIFF_VIEW(iter->userdata), 0);
 	g_return_val_if_fail(iter->userdata2 != NULL, 0);
-	
+
 	return ((Region *)iter->userdata2)->type;
 }
 
@@ -810,12 +795,12 @@ static void
 region_iter_range(GitgDiffView *view, Region *region, GtkTextIter *start, GtkTextIter *end)
 {
 	gtk_text_buffer_get_iter_at_line(view->priv->current_buffer, start, region->line);
-	
+
 	Region *next = region->next;
-	
+
 	while (next && next->type != region->type)
 		next = next->next;
-	
+
 	if (next)
 		gtk_text_buffer_get_iter_at_line(view->priv->current_buffer, end, next->line);
 	else
@@ -831,21 +816,21 @@ gitg_diff_iter_set_visible(GitgDiffIter *iter, gboolean visible)
 
 	GitgDiffView *view = GITG_DIFF_VIEW(iter->userdata);
 	Region *region = (Region *)iter->userdata2;
-	
+
 	if (region->visible == visible)
 		return;
-	
+
 	GtkTextIter start;
 	GtkTextIter end;
-	
+
 	region_iter_range(view, region, &start, &end);
 	region->visible = visible;
-	
+
 	/* Propagate visibility to hunks */
 	if (region->type == GITG_DIFF_ITER_TYPE_HEADER)
 	{
 		Region *next = region->next;
-		
+
 		while (next && next->type != GITG_DIFF_ITER_TYPE_HEADER)
 		{
 			next->visible = visible;
@@ -856,11 +841,21 @@ gitg_diff_iter_set_visible(GitgDiffIter *iter, gboolean visible)
 	if (visible)
 	{
 		gtk_text_buffer_remove_tag(view->priv->current_buffer, view->priv->invisible_tag, &start, &end);
+
+		if (region->type == GITG_DIFF_ITER_TYPE_HEADER)
+		{
+			hide_header_details (view, region);
+		}
 	}
 	else
 	{
+		if (region->type == GITG_DIFF_ITER_TYPE_HEADER)
+		{
+			gtk_text_buffer_remove_tag (view->priv->current_buffer, view->priv->subheader_tag, &start, &end);
+		}
+
 		gtk_text_buffer_apply_tag(view->priv->current_buffer, view->priv->invisible_tag, &start, &end);
-	}	
+	}
 }
 
 static gboolean
@@ -875,14 +870,14 @@ header_parse_index(GitgDiffView *view, Header *header)
 		num = header->region.next->line - header->region.line;
 	else
 		num = gtk_text_buffer_get_line_count(buffer) - header->region.line;
-	
+
 	gtk_text_buffer_get_iter_at_line(buffer, &iter, header->region.line);
-	
+
 	for (i = 0; i < num; ++i)
 	{
 		if (!gtk_text_iter_forward_line(&iter))
 			return FALSE;
-		
+
 		GtkTextIter end = iter;
 		gtk_text_iter_forward_to_line_end(&end);
 
@@ -895,18 +890,18 @@ header_parse_index(GitgDiffView *view, Header *header)
 			gchar *start = line + strlen(match);
 			gchar *sep = strstr(start, "..");
 			gboolean ret;
-			
+
 			if (sep)
 			{
 				gchar *last = strstr(sep, " ");
 				gchar *bet = strstr(start, ",");
-				
+
 				if (!last)
 					last = line + strlen(line);
 
 				strncpy(header->index_from, start, (bet ? bet : sep) - start);
 				strncpy(header->index_to, sep + 2, last - (sep + 2));
-				
+
 				header->index_from[(bet ? bet : sep) - start] = '\0';
 				header->index_to[last - (sep + 2)] = '\0';
 
@@ -916,14 +911,14 @@ header_parse_index(GitgDiffView *view, Header *header)
 			{
 				ret = FALSE;
 			}
-			
+
 			g_free(line);
 			return ret;
 		}
-		
+
 		g_free(line);
 	}
-	
+
 	return FALSE;
 }
 
@@ -931,25 +926,25 @@ gboolean
 gitg_diff_iter_get_index(GitgDiffIter *iter, gchar **from, gchar **to)
 {
 	Region *region = (Region *)iter->userdata2;
-	
+
 	while (region && region->type != GITG_DIFF_ITER_TYPE_HEADER)
 		region = region->prev;
-	
+
 	if (!region)
 		return FALSE;
-	
+
 	Header *header = (Header *)region;
 	gboolean ret = TRUE;
-	
+
 	if (!*(header->index_to))
 		ret = header_parse_index((GitgDiffView *)iter->userdata, header);
-	
+
 	if (!ret)
 		return FALSE;
-	
+
 	*from = header->index_from;
 	*to = header->index_to;
-	
+
 	return TRUE;
 }
 
@@ -960,11 +955,11 @@ iter_in_view(GitgDiffView *view, GtkTextIter *iter)
 	GtkTextIter end;
 	GdkRectangle rect;
 	GtkTextView *textview = GTK_TEXT_VIEW(view);
-	
+
 	gtk_text_view_get_visible_rect(textview, &rect);
 	gtk_text_view_get_iter_at_location(textview, &start, rect.x, rect.y);
 	gtk_text_view_get_iter_at_location(textview, &end, rect.x + rect.width, rect.y + rect.height);
-	
+
 	return gtk_text_iter_in_range(iter, &start, &end) || gtk_text_iter_equal(iter, &end);
 }
 
@@ -972,12 +967,12 @@ static gboolean
 try_scan(GitgDiffView *view)
 {
 	gint lines = gtk_text_buffer_get_line_count(view->priv->current_buffer);
-	
+
 	if (view->priv->last_scan_line > lines)
 		return FALSE;
 
 	guint num = MIN(lines - view->priv->last_scan_line, IDLE_SCAN_COUNT);
-	
+
 	if (num == 0)
 		return FALSE;
 
@@ -988,10 +983,10 @@ try_scan(GitgDiffView *view)
 	guint last = view->priv->last_scan_line;
 	ensure_scan(view, view->priv->last_scan_line + num);
 	g_snprintf(str, sizeof(str), "%u", view->priv->max_line_count);
-	
+
 	if (strlen(str) > max_line)
 		gtk_widget_queue_draw(GTK_WIDGET(view));
-	
+
 	return last != view->priv->last_scan_line;
 }
 
@@ -1023,7 +1018,7 @@ on_idle_scan(GitgDiffView *view)
 {
 	if (try_scan(view))
 		return TRUE;
-	
+
 	view->priv->scan_id = 0;
 	return FALSE;
 }
